from pydantic import BaseModel, Field
from typing import List, Dict, Any


class TaskConfigurationSchema(BaseModel):
    task_input: List[Any] = Field(title="task_input",
                                  description="The input for the task",
                                  default=[])
    detail: List[Dict] = Field(title="detail",
                               description="Details pertaining to the task",
                               default=[])

    task_packages_versions_required: Dict = Field(title="task_packages_versions_required",
                                                  description="Specific packages versions the task requires",
                                                  default={})

    stage_task_id: int = Field(title="stage_task_id",
                               description="The id of the stage that this task has. "
                                           "By default it is assumed that the task is not part of any StagedTask",
                               default=-1)

    has_next_stage: bool = Field(title="has_next_stage",
                                 description="Flag indicating if the task has a next stage. "
                                             "By default it is assumed that the task is not part of any StagedTask",
                                 default=False)

    next_stage_id: int = Field(title="next_stage_id",
                               description="If the task has next stage then this is id of the next stage. "
                                           "By default it is assumed that the task is not part of any StagedTask",
                               default=-1)



import bson
from navalmartin_mir_db_utils.utils.exceptions import InvalidObjectIdException

INVALID_BSON_ID = "INVALID_BSON_OBJ_ID"

DB_INFO = "DB_INFO: "
DB_WARNING = "DB_WARNING: "
DB_ERROR = "DB_ERROR: "


def is_valid_object_id(oid: str) -> bson.ObjectId:
    """Attempts to change the given oid in a valid bson.ObjectId.
    It raises InvalidObjectIdException if the cast fials
    Paramaters
    ----------
    oid: The object id to check

    Returns
    -------

    An instance of bson.ObjectId upon successfully casting the oid in 
    a bson.ObjectId
    """
    try:
        oid = bson.ObjectId(oid)
        return oid
    except Exception as e:
        raise InvalidObjectIdException(oid=oid)



"""module db_enums. Various useful enumerations

"""
from enum import Enum


class InvalidTypeEnum(Enum):
    INVALID = 0


class TaskStateTypeEnum(Enum):
    INVALID = 0
    CANCELLED = 1
    IN_PROGRESS = 2
    PENDING = 3
    FINISHED = 4
    FINISHED_WITH_ERRORS = 5


class ErrorTypeEnum(Enum):
    UNKNOWN_ERROR = 0
    GENERAL_SQS_MESSAGE_ERROR = 1
    GENERAL_DATA_ERROR = 2
    FRAMEWORK_EXCEPTION = 3
    INCORRECT_RESULT = 4



from .transform_utils import (stringify_dictionary, stringify_list)
from .exceptions import (InvalidObjectIdException,
                         ResourceNotFoundException,
                         ResourceNotUpdatedException,
                         ResourceExistsException,
                         DBInsertFailedException,
                         DBUpdateFailedException,
                         InvalidAttributeValueException,
                         InvalidMongoDBOperatorException)
from .db_enums import (TaskStateTypeEnum, ErrorTypeEnum, InvalidTypeEnum)



from typing import List, Any


def stringify_dictionary(data: dict, names: List[str]) -> dict:
    for name in data:
        if name in names:
            data[name] = str(data[name])
    return data


def stringify_list(data: List[Any]) -> List[str]:
    data = [str(item) for item in data]
    return data


from pydantic import BaseModel, Field, EmailStr
from typing import Optional
from navalmartin_mir_db_utils.schemata.data_views.indexed_item_data_view_base import IndexedItemDataViewBase


class UserDataViewBase(BaseModel):
    idx: Optional[str] = Field(title="idx",
                               description="The MongoDB index of the item",
                               alias="_id")

    email: Optional[EmailStr] = Field(
        title="email", description="The email of the user"
    )
    name: Optional[str] = Field(title="name", 
                                description="The name of the user")
    surname: Optional[str] = Field(
        title="surname", description="The surname of the user"
    )

    created_at: Optional[str] = Field(
        title="created_at", description="Time and date the particular entry was created"
    )
    updated_at: Optional[str] = Field(
        title="updated_at", description="Time and date the particular entry was updated"
    )

    access_token: Optional[str] = Field(title="access_token", description="The access token assigned to the user")
    refresh_token: Optional[str] = Field(title="refresh_token", description="The access token assigned to the user")

    @staticmethod
    def build_from_mongodb_json(
        mdb_json: dict,
        access_token: str,
        refresh_token: str,
    ) -> "UserDataViewBase":
        """Build the user data from the specified MongoDB document

        Parameters
        ----------
        mdb_json: The MongoDB document
        access_token: The access token
        refresh_token: The refresh token

        Returns
        -------

        """

        mdb_json["index_data"] = IndexedItemDataViewBase.build_from_mongodb_json(mdb_json=mdb_json).dict()

        mdb_json["access_token"] = access_token
        mdb_json["refresh_token"] = refresh_token

        return UserDataViewBase(**mdb_json)



from setuptools import setup

setup(
    name='navalmartin_mir_db_utils',
    packages=['navalmartin_mir_db_utils'],
    version='0.0.1',
    description='Various utilities for working with AWS in Python 3.',
    author='Alexandros Giavaras',
    author_email='a.giavaras@gmail.com',
    url='https://github.com/Navalmartin/mir_persistence_layer_utils',
    download_url='https://github.com/pockerman/awsutils/awsutils-0.0.3.tar.gz',
    keywords=['mongodb', 'mongodb-utilities'],
    classifiers=[],
    scripts=[],
)



from typing import Any, Callable
from pymongo.results import InsertOneResult, UpdateResult

from navalmartin_mir_db_utils.dbs.dbs_utils import DB_ERROR
from navalmartin_mir_db_utils.crud.mongodb_crud_ops import ReadEntityCRUDAPI, CreateEntityCRUDAPI, UpdateEntityCRUDAPI
from navalmartin_mir_db_utils.dbs.mongodb_session import MongoDBSession
from navalmartin_mir_db_utils.utils.exceptions import (ResourceNotFoundException, ResourceNotUpdatedException,
                                                       ResourceExistsException, DBInsertFailedException)


async def get_one_result_or_raise(
        crud_handler: ReadEntityCRUDAPI,
        criteria: dict,
        db_session: MongoDBSession,
        projection: dict = {},
        error_message: str = "Error occurred"
) -> Any:
    """Query the handler about a result and raise
    ResourceNotFoundException if the query does not return any
    result
    """
    result = await crud_handler.find_one(
        criteria=criteria, db_session=db_session, projection=projection,
        collection_name=crud_handler.collection_name
    )

    if result is None:
        print(f"{DB_ERROR} {error_message}")
        raise ResourceNotFoundException(search_criteria=str(criteria))

    return result


async def insert_one_or_fail(crud_handler: CreateEntityCRUDAPI,
                             insert_data: dict,
                             db_session: MongoDBSession,
                             error_message: str = "Error occurred",
                             schema: Callable = None) -> InsertOneResult:
    """

    Parameters
    ----------
    crud_handler: The handler that does the insert
    insert_data: The data to insert
    db_session: The session used
    error_message: The error message to show
    schema: The schema to validate the data
    Returns
    -------

    An instance of InsertOneResult
    """

    if schema is not None:
        try:
            schema(**insert_data)
        except Exception as e:
            print(f"{DB_ERROR} Schema validation failed on insert")
            raise DBInsertFailedException(collection_name=crud_handler.collection_name)

    result: InsertOneResult = await crud_handler.insert_one(data=insert_data,
                                                            db_session=db_session,
                                                            collection_name=crud_handler.collection_name)

    if result is None:
        print(f"{DB_ERROR} {error_message}")
        raise DBInsertFailedException(collection_name=crud_handler.collection_name)

    if not result.acknowledged:
        print(f"{DB_ERROR} {error_message}")
        raise DBInsertFailedException(collection_name=crud_handler.collection_name)

    return result


async def if_resource_found_raise(crud_handler,
                                  criteria: dict,
                                  db_session: MongoDBSession,
                                  error_message: str = "Error occurred"):
    result = await crud_handler.find_one(
        criteria=criteria, db_session=db_session, projection={'_id': 1},
        collection_name=crud_handler.collection_name
    )

    if result is not None:
        print(f"{DB_ERROR} {error_message}")
        raise ResourceExistsException(resource_id=str(criteria))


async def update_one_or_raise(
        crud_handler: UpdateEntityCRUDAPI,
        criteria: dict,
        db_session,
        update_data: dict,
        update_op: str = "$set",
        error_message: str = "Error occurred",
) -> UpdateResult:
    result: UpdateResult = await crud_handler.update_one(
        criteria=criteria, db_session=db_session, update_data=update_data,
        collection_name=crud_handler.collection_name, update_op=update_op
    )

    if result is None:
        print(f"{DB_ERROR} {error_message}")
        raise ResourceNotUpdatedException(resource_id=str(criteria))

    if not result.acknowledged:
        print(f"{DB_ERROR} {error_message}")
        raise ResourceNotUpdatedException(resource_id=str(criteria))

    if result.modified_count != 1:
        print(f"{DB_ERROR} {error_message}")
        raise ResourceNotUpdatedException(resource_id=str(criteria))

    return result



from functools import wraps
from typing import Callable, Any
from pymongo.read_concern import ReadConcern
from pymongo.write_concern import WriteConcern
from pymongo.read_preferences import ReadPreference
from navalmartin_mir_db_utils.dbs.mongodb_session import MongoDBSession
from navalmartin_mir_db_utils.dbs.dbs_utils import DB_ERROR, DB_INFO


def with_transaction(fn: Callable) -> Any:
    """Simple decorator to mark the callable
    as executing a transaction

    Parameters
    ----------
    fn: The callable to wrap

    Returns
    -------

    """
    @wraps(fn)
    async def wrapper(*args, **kwargs):
        func_result = await fn(*args, **kwargs)
        return func_result

    return wrapper


def use_async_transaction(mdb_session: MongoDBSession,
                          async_callback: Callable,
                          callback_args: dict,
                          max_commit_time_ms: int,
                          read_concern: ReadConcern, write_concern: WriteConcern,
                          read_preference: ReadPreference,
                          with_log: bool = False,
                          with_transaction_result: bool = None,
                          transaction_result_handler: Callable = None) -> Any:
    """Decorator to wrap the given callable into a MongoDB transaction
    The callback function implements the actual read/write operations.
    This is similar to with_transaction but assumes async ops

    Parameters
    ----------
    mdb_session: The MongoDB session
    async_callback: The function callback to be used
    callback_args: Dictionary with arguments to pass in the callback
    max_commit_time_ms: The maximum commit time for the transaction
    read_concern: The ReadConcern
    write_concern: The  WriteConcern
    read_preference: The ReadPreference
    with_log: If True it prints information massages on std output
    with_transaction_result: If True the transaction result is added in the
    kwargs that is sent to the function using this decorator
    transaction_result_handler: Callable that specifies how to handle the transaction result
    Returns
    -------

    """

    def _execute_func(async_fn: Callable):
        @wraps(async_fn)
        async def wrapper(*args, **kwargs):

            if with_log:
                print(f"{DB_INFO} Starting transaction")

            # Start a client session.
            async with await mdb_session.client.start_session() as session:

                try:
                    # Use with_transaction to start a transaction,
                    # execute the callback, and commit (or abort on error).
                    transaction_result = await session.with_transaction(
                        lambda s: async_callback(session, callback_args),
                        read_concern=read_concern,
                        write_concern=write_concern,
                        read_preference=read_preference,
                        max_commit_time_ms=max_commit_time_ms
                    )

                    if transaction_result_handler is not None:
                        transaction_result = await transaction_result_handler(transaction_result)

                    if with_transaction_result:
                        kwargs['transaction_result'] = transaction_result

                except Exception as e:
                    if with_log:
                        print(f"{DB_ERROR} Session with transaction raised an exception. Exception message={str(e)}")
                    raise e

            if with_log:
                print(f"{DB_INFO} End transaction")

            func_result = await async_fn(*args, **kwargs)
            return func_result

        return wrapper

    return _execute_func
